/*
 * Copyright (C) 2011 4th Line GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.fourthline.sash.mediaserver;

import android.app.Notification;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.app.Service;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.net.wifi.WifiManager;
import android.os.IBinder;
import org.fourthline.android.AndroidLocalInetAddressResolver;
import org.fourthline.android.LocalInetAddressResolver;
import org.fourthline.android.httpserver.HttpServerService;
import org.fourthline.android.httpserver.HttpServerServiceImpl;
import org.fourthline.cling.android.AndroidUpnpService;
import org.fourthline.cling.android.AndroidUpnpServiceImpl;
import org.fourthline.cling.model.meta.LocalDevice;
import org.fourthline.cling.support.model.DIDLObject;
import org.fourthline.sash.MainActivity;
import org.fourthline.sash.content.MediaStoreContent;
import org.fourthline.sash.content.URLBuilder;
import org.seamless.util.URIUtil;

import java.net.URI;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This Android service access/uses two other Android services:
 * <p>
 * The first is the HTTP server for providing actual media content (the media data).
 * The second is the UPnP server, which provides the metadata about the media, using
 * a separate HTTP (internal) HTTP server.
 * </p>
 *
 * @author Christian Bauer
 */
public class MediaServerServiceImpl extends Service {

    final private static Logger log = Logger.getLogger(MediaServerService.class.getName());

    protected Binder binder = new Binder();

    protected class Binder extends android.os.Binder implements MediaServerService {

    }

    private static final int NOTIFICATION_ID = 1;

    protected ContentHttpServerConnection contentHttpServerConnection;
    protected UpnpServiceConnection upnpServiceConnection;

    @Override
    public void onCreate() {
        super.onCreate();

        getNotificationManager().notify(NOTIFICATION_ID, createNotification());

        contentHttpServerConnection = new ContentHttpServerConnection(
                this,
                new AndroidLocalInetAddressResolver((WifiManager) getSystemService(Context.WIFI_SERVICE))
        );

        upnpServiceConnection = new UpnpServiceConnection(
                new MediaServer(contentHttpServerConnection.getContent())
        );

        contentHttpServerConnection.getContent().registerObservers();
        contentHttpServerConnection.getContent().updateAll();

        log.fine("Binding to content HTTP server service");
        getApplicationContext().bindService(
                new Intent(this, HttpServerServiceImpl.class),
                contentHttpServerConnection,
                Context.BIND_AUTO_CREATE
        );

        log.fine("Binding to UPnP service");
        getApplicationContext().bindService(
                new Intent(MediaServerServiceImpl.this, AndroidUpnpServiceImpl.class),
                upnpServiceConnection,
                Context.BIND_AUTO_CREATE
        );
    }

    @Override
    public void onDestroy() {
        super.onDestroy();

        if (contentHttpServerConnection != null) {

            contentHttpServerConnection.getContent().unregisterObservers();

            log.fine("Unbinding from HTTP server service");
            contentHttpServerConnection.prepareUnbind();
            getApplicationContext().unbindService(contentHttpServerConnection);
        }

        if (upnpServiceConnection != null) {
            log.fine("Unbinding from UPnP service");
            upnpServiceConnection.prepareUnbind();
            getApplicationContext().unbindService(upnpServiceConnection);
        }

        getNotificationManager().cancel(NOTIFICATION_ID);
    }

    @Override
    public IBinder onBind(Intent intent) {
        return binder;
    }

    protected NotificationManager getNotificationManager() {
        return (NotificationManager) getSystemService(Context.NOTIFICATION_SERVICE);
    }

    protected Notification createNotification() {
        Notification notification = new Notification();

        notification.icon = android.R.drawable.stat_notify_sync_noanim;
        notification.tickerText = "Sash MediaServer is active";

        notification.flags |= Notification.FLAG_ONGOING_EVENT;
        notification.flags |= Notification.FLAG_NO_CLEAR;

        notification.setLatestEventInfo(
                getApplicationContext(),
                notification.tickerText,
                null,
                PendingIntent.getActivity(this, 0, new Intent(this, MainActivity.class), 0));

        return notification;
    }

    // The adapter for the content HTTP server
    class ContentHttpServerConnection implements ServiceConnection {

        final protected MediaStoreContent content;
        protected HttpServerService httpServerService;

        ContentHttpServerConnection(Context context, final LocalInetAddressResolver localAddressResolver) {

            log.fine("Creating MediaStore content");
            content = new MediaStoreContent(
                    context,
                    new URLBuilder() {
                        public String getURL(DIDLObject object) {
                            return URIUtil.createAbsoluteURL(
                                    localAddressResolver.getLocalInetAddress(),
                                    getHttpServerService().getLocalPort(),
                                    URI.create("/" + object.getId())
                            ).toString();
                        }

                        public String getObjectId(String urlPath) {
                            return urlPath.substring(1); // Cut the slash
                        }
                    }
            );
        }

        public MediaStoreContent getContent() {
            return content;
        }

        public HttpServerService getHttpServerService() {
            return httpServerService;
        }

        public void onServiceConnected(ComponentName componentName, IBinder service) {
            httpServerService = (HttpServerService) service;
            httpServerService.addHandler("*", content);
        }

        public void onServiceDisconnected(ComponentName componentName) {
            httpServerService = null;
        }

        public void prepareUnbind() {
            if (httpServerService != null) {
                httpServerService.removeHandler("*");
            }
        }
    }

    // The adapter for the UPnP service
    class UpnpServiceConnection implements ServiceConnection {

        final protected MediaServer mediaServer;
        protected AndroidUpnpService upnpService;

        UpnpServiceConnection(MediaServer mediaServer) {
            this.mediaServer = mediaServer;
        }

        public void onServiceConnected(ComponentName className, IBinder service) {
            upnpService = (AndroidUpnpService) service;

            LocalDevice mediaServerDevice =
                    upnpService.getRegistry().getLocalDevice(mediaServer.getUdn(), true);
            if (mediaServerDevice == null) {
                try {
                    log.fine("Creating MediaServer device and registering with UPnP service");
                    mediaServerDevice = mediaServer.createDevice();
                    upnpService.getRegistry().addDevice(mediaServerDevice);
                } catch (Exception ex) {
                    log.log(Level.SEVERE, "Creating or registering media server device failed", ex);
                }
            }
        }

        public void onServiceDisconnected(ComponentName className) {
            upnpService = null;
        }

        public void prepareUnbind() {
            if (upnpService != null) {
                upnpService.getRegistry().removeDevice(mediaServer.getUdn());
            }
        }
    }
}
